
#include <nds.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "_console.h"
#include "_consolewritelog.h"
#include "maindef.h"
#include "memtool.h"
#include "_const.h"
#include "lang.h"
#include "inifile.h"
#include "strtool.h"

#include "glib/glib.h"

#include "fat2.h"
#include "zlibhelp.h"
#include "cstream_fs.h"
#include "cipk.h"
#include "ipkfilelist.h"
#include "ipkstate.h"
#include "corner.h"

#include "rect.h"
#include "component.h"

#include "proc_screen.h"

#define BGColor15 (RGB15(25,22,23) | BIT15)

static TIPKState_SlideSetting SlideSetting;

enum ECompLabels {ECLS_TitleLbl,ECLS_DetectTtlLbl,
                  ECLS_DetectGlobalFileLbl,ECLS_DetectGlobalFileExistLbl,
                  ECLS_DetectGlobalPathLbl,ECLS_DetectGlobalPathExistLbl,
                  ECLS_DetectEqualFilenameLbl,ECLS_DetectEqualFilenameExistLbl,
                  ECLS_DetectEqualFolderLbl,ECLS_DetectEqualFolderExistLbl,
                  ECLS_SelectedBGMLbl,ECLS_BGMSelectLbl,ECLSCount};
#define CompLabelsCount (ECLSCount)
static TComponentLabel CompLabels[CompLabelsCount];

enum ECompChecks {ECCS_BGMSelectSequenceRadio,ECCS_BGMSelectShuffleRadio,ECCSCount};
#define CompChecksCount (ECCSCount)
static TComponentCheck CompChecks[CompChecksCount];

enum ECompButtons {ECBS_Page1Btn,ECBS_Page2Btn,ECBS_Page3Btn,ECBS_Page4Btn,ECBS_StartBtn,ECBS_CancelBtn,ECBSCount};
#define CompButtonsCount (ECBSCount)
static TComponentButton CompButtons[CompButtonsCount];

static void SlideSetting_Redraw(bool FullDraw)
{
  if(FullDraw==true){
    CglCanvas *pcan=pScreenSubBM->pCanvas;
    
    u32 TitleHeight=30;
    pcan->SetColor(GlobalBGColor15);
    pcan->FillFast(0,0,ScreenWidth,TitleHeight);
    pcan->SetColor(BGColor15);
    pcan->FillFast(0,TitleHeight,ScreenWidth,ScreenHeight-TitleHeight);
  }
  
  TIPKState_SlideSetting *pss=&SlideSetting;
  
  {
    CompChecks[ECCS_BGMSelectSequenceRadio].Checked=false;
    CompChecks[ECCS_BGMSelectShuffleRadio].Checked=false;
    
    switch(pss->BGMSelect){
      case EBS_Sequence: CompChecks[ECCS_BGMSelectSequenceRadio].Checked=true; break;
      case EBS_Shuffle: CompChecks[ECCS_BGMSelectShuffleRadio].Checked=true; break;
      default: break;
    }
  }
  
  if(FullDraw==true){
    for(u32 idx=0;idx<CompLabelsCount;idx++){
      ComponentLabel_Draw(&CompLabels[idx]);
    }
    for(u32 idx=0;idx<CompChecksCount;idx++){
      ComponentCheck_Draw(&CompChecks[idx]);
    }
    for(u32 idx=0;idx<CompButtonsCount;idx++){
      ComponentButton_Draw(&CompButtons[idx]);
    }
    }else{
    ComponentCheck_Draw(&CompChecks[ECCS_BGMSelectSequenceRadio]);
    ComponentCheck_Draw(&CompChecks[ECCS_BGMSelectShuffleRadio]);
  }
}

static void CB_BGMSelectSequenceRadio_Click(void *pComponentCheck)
{
  TIPKState_SlideSetting *pss=&SlideSetting;
  pss->BGMSelect=EBS_Sequence;
  SlideSetting_Redraw(false);
}

static void CB_BGMSelectShuffleRadio_Click(void *pComponentCheck)
{
  TIPKState_SlideSetting *pss=&SlideSetting;
  pss->BGMSelect=EBS_Shuffle;
  SlideSetting_Redraw(false);
}

static void CB_Page1Btn_Click(void *pComponentButton)
{
  TIPKState_SlideSetting *pss=&SlideSetting;
  pss->SlideOptPage=1;
  NextProc=ENP_SlideOpt1;
}

static void CB_Page2Btn_Click(void *pComponentButton)
{
  TIPKState_SlideSetting *pss=&SlideSetting;
  pss->SlideOptPage=2;
  NextProc=ENP_SlideOpt2;
}

static void CB_Page3Btn_Click(void *pComponentButton)
{
  TIPKState_SlideSetting *pss=&SlideSetting;
  pss->SlideOptPage=3;
  NextProc=ENP_SlideOpt3;
}

static void CB_Page4Btn_Click(void *pComponentButton)
{
  TIPKState_SlideSetting *pss=&SlideSetting;
  pss->SlideOptPage=4;
  NextProc=ENP_SlideOpt4;
}

static void CB_StartBtn_Click(void *pComponentButton)
{
  NextProc=ENP_SlidePlay;
}

static void CB_CancelBtn_Click(void *pComponentButton)
{
  NextProc=ENP_Thumbnail;
}

static void CompsInit(void)
{
  CglCanvas *pcan=pScreenSubBM->pCanvas;
  
  for(u32 idx=0;idx<CompLabelsCount;idx++){
    ComponentLabel_Init(&CompLabels[idx],pcan);
    CompLabels[idx].BGColor=BGColor15;
  }
  for(u32 idx=0;idx<CompChecksCount;idx++){
    ComponentCheck_Init(&CompChecks[idx],pcan);
    CompChecks[idx].BGColor=BGColor15;
  }
  for(u32 idx=0;idx<CompButtonsCount;idx++){
    ComponentButton_Init(&CompButtons[idx],pcan);
    CompButtons[idx].BGColor=BGColor15;
  }
  
  s32 chksize=16;
  
  s32 x,y;
  
  x=ScreenWidth-68-2;
  y=8-2;
  
  {
    TComponentButton *pcb=&CompButtons[ECBS_Page1Btn];
    pcb->CallBack_Click=CB_Page1Btn_Click;
    pcb->Rect=CreateRect(x,y,14,chksize+2);
    pcb->pMsgUTF8_eng="1";
    pcb->pMsgUTF8_jpn="1";
  }
  
  x+=18;
  
  {
    TComponentButton *pcb=&CompButtons[ECBS_Page2Btn];
    pcb->CallBack_Click=CB_Page2Btn_Click;
    pcb->Rect=CreateRect(x,y,14,chksize+2);
    pcb->pMsgUTF8_eng="2";
    pcb->pMsgUTF8_jpn="2";
  }
  
  x+=18;
  
  {
    TComponentButton *pcb=&CompButtons[ECBS_Page3Btn];
    pcb->CallBack_Click=CB_Page3Btn_Click;
    pcb->Rect=CreateRect(x,y,14,chksize+2);
    pcb->pMsgUTF8_eng="3";
    pcb->pMsgUTF8_jpn="3";
    pcb->Pressing=true;
  }
  
  x+=18;
  
  {
    TComponentButton *pcb=&CompButtons[ECBS_Page4Btn];
    pcb->CallBack_Click=CB_Page4Btn_Click;
    pcb->Rect=CreateRect(x,y,14,chksize+2);
    pcb->pMsgUTF8_eng="4";
    pcb->pMsgUTF8_jpn="4";
  }
  
  x=4;
  y=8;
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_TitleLbl];
    pcl->BGColor=GlobalBGColor15;
    pcl->Rect=CreateRect(x,y,0,0);
    pcl->pMsgUTF8_eng="BGM setting";
    pcl->pMsgUTF8_jpn="BGM設定";
  }
  
  x=12;
  y+=14+8+16;
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_DetectTtlLbl];
    pcl->Rect=CreateRect(x,y,0,0);
    pcl->pMsgUTF8_eng="Detected BGM files.";
    pcl->pMsgUTF8_jpn="検出されたBGMファイル";
  }
  
  x+=12;
  y+=20;
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_DetectGlobalFileLbl];
    pcl->Rect=CreateRect(x,y,0,0);
    pcl->pMsgUTF8_eng="imgview.mp3";
    pcl->pMsgUTF8_jpn="imgview.mp3";
  }
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_DetectGlobalFileExistLbl];
    pcl->Rect=CreateRect(x+140,y,0,0);
    if(IPK_GetGlobalBGMFilename()[0]!=0){
      pcl->TextColor=RGB15(0,0,15)|BIT15;
      pcl->pMsgUTF8_eng="OK";
      }else{
      pcl->TextColor=RGB15(15,0,0)|BIT15;
      pcl->pMsgUTF8_eng="NG";
    }
    pcl->pMsgUTF8_jpn=pcl->pMsgUTF8_eng;
  }
  
  y+=16;
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_DetectGlobalPathLbl];
    pcl->Rect=CreateRect(x,y,0,0);
    if(IPK_GetGlobalBGMFilePath()[0]!=0){
      pcl->pMsgUTF8_eng=IPK_GetGlobalBGMFilePath();
      pcl->pMsgUTF8_jpn=IPK_GetGlobalBGMFilePath();
      }else{
      pcl->pMsgUTF8_eng="folder from INI";
      pcl->pMsgUTF8_jpn="INI設定フォルダ";
    }
  }
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_DetectGlobalPathExistLbl];
    pcl->Rect=CreateRect(x+140,y,0,0);
    if(IPK_GetGlobalBGMFilePath()[0]!=0){
      pcl->TextColor=RGB15(0,0,15)|BIT15;
      pcl->pMsgUTF8_eng="OK";
      }else{
      pcl->TextColor=RGB15(15,0,0)|BIT15;
      pcl->pMsgUTF8_eng="NG";
    }
    pcl->pMsgUTF8_jpn=pcl->pMsgUTF8_eng;
  }
  
  y+=16;
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_DetectEqualFilenameLbl];
    pcl->Rect=CreateRect(x,y,0,0);
    pcl->pMsgUTF8_eng="Equal filename.mp3";
    pcl->pMsgUTF8_jpn="同ファイル名.mp3";
  }
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_DetectEqualFilenameExistLbl];
    pcl->Rect=CreateRect(x+140,y,0,0);
    if(IPK_GetBGMFilename(CurrentIPKFileIndex)[0]!=0){
      pcl->TextColor=RGB15(0,0,15)|BIT15;
      pcl->pMsgUTF8_eng="OK";
      }else{
      pcl->TextColor=RGB15(15,0,0)|BIT15;
      pcl->pMsgUTF8_eng="NG";
    }
    pcl->pMsgUTF8_jpn=pcl->pMsgUTF8_eng;
  }
  
  y+=16;
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_DetectEqualFolderLbl];
    pcl->Rect=CreateRect(x,y,0,0);
    pcl->pMsgUTF8_eng="Equal folder/*.mp3"; /* */
    pcl->pMsgUTF8_jpn="同フォルダ名/*.mp3"; /* */
  }
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_DetectEqualFolderExistLbl];
    pcl->Rect=CreateRect(x+140,y,0,0);
    if(IPK_GetBGMPath(CurrentIPKFileIndex)[0]!=0){
      pcl->TextColor=RGB15(0,0,15)|BIT15;
      pcl->pMsgUTF8_eng="OK";
      }else{
      pcl->TextColor=RGB15(15,0,0)|BIT15;
      pcl->pMsgUTF8_eng="NG";
    }
    pcl->pMsgUTF8_jpn=pcl->pMsgUTF8_eng;
  }
  
  x-=12;
  y+=20;
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_SelectedBGMLbl];
    pcl->Rect=CreateRect(x,y,0,0);
    switch(IPK_GetBGMType(CurrentIPKFileIndex)){
      case EBT_None: {
        pcl->pMsgUTF8_eng="Not found BGM Files.";
        pcl->pMsgUTF8_jpn="ファイルが見つかりません。";
      } break;
      case EBT_GlobalFile: {
        pcl->pMsgUTF8_eng="Play imgview.mp3.";
        pcl->pMsgUTF8_jpn="imgview.mp3を再生";
      } break;
      case EBT_GlobalPath: {
        static char eng[32];
        StrCopy("Play ",eng);
        StrAppend(eng,IPK_GetGlobalBGMFilePath());
        pcl->pMsgUTF8_eng=eng;
        static char jpn[32];
        StrCopy(IPK_GetGlobalBGMFilePath(),jpn);
        StrAppend(jpn," を再生");
        pcl->pMsgUTF8_jpn=jpn;
      } break;
      case EBT_File: {
        pcl->pMsgUTF8_eng="Play equal filename.";
        pcl->pMsgUTF8_jpn="BGMファイルを再生";
      } break;
      case EBT_Path: {
        pcl->pMsgUTF8_eng="Play equal folder.";
        pcl->pMsgUTF8_jpn="BGMフォルダを再生";
      } break;
      default: {
        pcl->pMsgUTF8_eng="BGMType error.";
        pcl->pMsgUTF8_jpn="BGMTypeエラー";
      }
    }
  }
  
  u16 BGMSelectColor=GlobalTextColor15;
  if(IPK_GetBGMType(CurrentIPKFileIndex)!=EBT_Path) BGMSelectColor=RGB15(20,18,18)|BIT15;
  
  y+=28;
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_BGMSelectLbl];
    pcl->Rect=CreateRect(x,y,0,0);
    pcl->TextColor=BGMSelectColor;
    pcl->pMsgUTF8_eng="Folder playback mode";
    pcl->pMsgUTF8_jpn="フォルダ再生モード";
  }
  
  x+=12;
  y+=20;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_BGMSelectSequenceRadio];
    pcc->CallBack_Click=CB_BGMSelectSequenceRadio_Click;
    pcc->pOnCanvas=Component_GetIcon(ECIT_RadioOn)->pCanvas;
    pcc->pOffCanvas=Component_GetIcon(ECIT_RadioOff)->pCanvas;
    pcc->TextColor=BGMSelectColor;
    pcc->pMsgUTF8_eng="Sequence loop";
    pcc->pMsgUTF8_jpn="順ループ";
    pcc->Rect=CreateRect(x,y,chksize,chksize);
  }
  
  y+=20;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_BGMSelectShuffleRadio];
    pcc->CallBack_Click=CB_BGMSelectShuffleRadio_Click;
    pcc->pOnCanvas=Component_GetIcon(ECIT_RadioOn)->pCanvas;
    pcc->pOffCanvas=Component_GetIcon(ECIT_RadioOff)->pCanvas;
    pcc->TextColor=BGMSelectColor;
    pcc->pMsgUTF8_eng="Shuffle loop";
    pcc->pMsgUTF8_jpn="シャッフルループ";
    pcc->Rect=CreateRect(x,y,chksize,chksize);
  }
  
  x-=12;
  
  x=(ScreenWidth-((80*2)+16))/2;
  y=ScreenHeight-(chksize+2)-8;
  
  {
    TComponentButton *pcb=&CompButtons[ECBS_StartBtn];
    pcb->CallBack_Click=CB_StartBtn_Click;
    pcb->pIconCanvas=Component_GetIcon(ECIT_Ok)->pCanvas;
    pcb->Rect=CreateRect(x,y,80,chksize+2);
    pcb->pMsgUTF8_eng="Start";
    pcb->pMsgUTF8_jpn="開始";
  }
  
  x+=16+80;
  
  {
    TComponentButton *pcb=&CompButtons[ECBS_CancelBtn];
    pcb->CallBack_Click=CB_CancelBtn_Click;
    pcb->pIconCanvas=Component_GetIcon(ECIT_Cancel)->pCanvas;
    pcb->Rect=CreateRect(x,y,80,chksize+2);
    pcb->pMsgUTF8_eng="Cancel";
    pcb->pMsgUTF8_jpn="戻る";
  }
}

static void CB_KeyDown(u32 Keys)
{
}

static void CB_KeyPress(u32 VsyncCount,u32 Keys)
{
  if((Keys&KEY_START)!=0) CB_StartBtn_Click(NULL);
}

static void CB_KeyUp(u32 Keys)
{
}

static bool deskmf;
static TComponentButton *pPressingButton;

static void CB_MouseDown(s32 x,s32 y)
{
  deskmf=false;
  
  for(u32 idx=0;idx<CompButtonsCount;idx++){
    TComponentButton *pcb=&CompButtons[idx];
    if(ComponentButton_GetIndexFromPos(pcb,x,y)!=-1){
      pPressingButton=pcb;
      pcb->Pressing=true;
      ComponentButton_Draw(pcb);
    }
  }
  
  deskmf=true;
}

static void CB_MouseMove(s32 x,s32 y,bool CanIgnore)
{
  if(deskmf==false) return;
  if(CanIgnore==true) return;
  
  for(u32 idx=0;idx<CompButtonsCount;idx++){
    TComponentButton *pcb=&CompButtons[idx];
    if(pcb==pPressingButton){
      if(ComponentButton_GetIndexFromPos(pcb,x,y)==-1){
        if(pcb->Pressing==true){
          pcb->Pressing=false;
          ComponentButton_Draw(pcb);
        }
        }else{
        if(pcb->Pressing==false){
          pcb->Pressing=true;
          ComponentButton_Draw(pcb);
        }
      }
    }
  }
}

static void CB_MouseUp(s32 x,s32 y)
{
  if(deskmf==false) return;
  
  deskmf=false;
  
  if(pPressingButton!=NULL){
    pPressingButton->Pressing=false;
    ComponentButton_Draw(pPressingButton);
    
    for(u32 idx=0;idx<CompButtonsCount;idx++){
      TComponentButton *pcb=&CompButtons[idx];
      if(pcb==pPressingButton){
        if(ComponentButton_GetIndexFromPos(pcb,x,y)!=-1){
          ComponentButton_MouseUp(&CompButtons[idx],x,y);
        }
      }
    }
    
    pPressingButton=NULL;
    return;
  }
  
  for(u32 idx=0;idx<CompLabelsCount;idx++){
    ComponentLabel_MouseUp(&CompLabels[idx],x,y);
  }
  for(u32 idx=0;idx<CompChecksCount;idx++){
    ComponentCheck_MouseUp(&CompChecks[idx],x,y);
  }
  for(u32 idx=0;idx<CompButtonsCount;idx++){
  }
}

static void CB_Start(void)
{
  ScreenInit(ESS_Bitmap,EDM_Vertical);
  
  {
    const u32 s=CornerIconSize;
    const u32 x=ScreenWidth;
    const u32 y=ScreenHeight-(s*8);
    CglB15 *pB15=Corner_GetIcon(EIT_Mouse);
    pScreenSubBM->Mouse_Init(x,y+(s*0),pB15->pCanvas);
  }
  
  {
    CglCanvas *pcan=pScreenMain->pCanvas;
    pcan->SetColor(BGColor15);
    pcan->FillAll();
  }
  {
    CglCanvas *pcan=pScreenSubBM->pCanvas;
    pcan->SetColor(BGColor15);
    pcan->FillAll();
  }
  
  DrawSlideOptLeft(3);
  
  deskmf=false;
  pPressingButton=NULL;
  
  pScreenMain->SetWhiteOut(16);
  pScreenSubBM->SetAlpha(0);
  
  CompsInit();
  
  SlideSetting=IPKState.SlideSetting;
  
//  pScreenSubBM->SetAlpha(8); // for Debug.
  
  SlideSetting_Redraw(true);
  
  for(u32 lev=0;lev<=16;lev+=2){
    WaitForVBlank(true);
    pScreenMain->SetWhiteOut(16-lev);
    pScreenSubBM->SetAlpha(lev);
  }
}

static void CB_VsyncUpdate(u32 VsyncCount)
{
}

static void CB_End(void)
{
  for(u32 lev=0;lev<=16;lev+=2){
    WaitForVBlank(true);
    pScreenMain->SetWhiteOut(lev);
    pScreenSubBM->SetAlpha(16-lev);
  }
  
  IPKState.SlideSetting=SlideSetting;
  ScreenEnd(ESS_Bitmap);
}

void ProcSlideOpt3_SetCallBack(void)
{
  CallBack_Start=CB_Start;
  CallBack_VsyncUpdate=CB_VsyncUpdate;
  CallBack_End=CB_End;
  CallBack_KeyDown=CB_KeyDown;
  CallBack_KeyPress=CB_KeyPress;
  CallBack_KeyUp=CB_KeyUp;
  CallBack_MouseDown=CB_MouseDown;
  CallBack_MouseMove=CB_MouseMove;
  CallBack_MouseUp=CB_MouseUp;
//  GlobalINI.System.StartButtonFunction=ESSBF_ShortCut;
}

